// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
void DfgVisitor::visit(DfgAdd* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgAnd* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgArraySel* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgAstRd* vtxp) { visit(static_cast<DfgVertexAst*>(vtxp)); }
void DfgVisitor::visit(DfgCReset* vtxp) { visit(static_cast<DfgVertexNullary*>(vtxp)); }
void DfgVisitor::visit(DfgConcat* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgCond* vtxp) { visit(static_cast<DfgVertexTernary*>(vtxp)); }
void DfgVisitor::visit(DfgConst* vtxp) { visit(static_cast<DfgVertexNullary*>(vtxp)); }
void DfgVisitor::visit(DfgCountOnes* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgDiv* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgDivS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgEq* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgEqCase* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgEqWild* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgExtend* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgExtendS* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgGt* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgGtS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgGte* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgGteS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgLogAnd* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgLogEq* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgLogIf* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgLogNot* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgLogOr* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgLogic* vtxp) { visit(static_cast<DfgVertexVariadic*>(vtxp)); }
void DfgVisitor::visit(DfgLt* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgLtS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgLte* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgLteS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgMatchMasked* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgModDiv* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgModDivS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgMul* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgMulS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgMux* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgNegate* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgNeq* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgNeqCase* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgNeqWild* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgNot* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgOneHot* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgOneHot0* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgOr* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgPow* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgPowSS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgPowSU* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgPowUS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgRedAnd* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgRedOr* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgRedXor* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgRep* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgSel* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgShiftL* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgShiftR* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgShiftRS* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgSpliceArray* vtxp) { visit(static_cast<DfgVertexSplice*>(vtxp)); }
void DfgVisitor::visit(DfgSplicePacked* vtxp) { visit(static_cast<DfgVertexSplice*>(vtxp)); }
void DfgVisitor::visit(DfgStreamL* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgStreamR* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgSub* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
void DfgVisitor::visit(DfgUnitArray* vtxp) { visit(static_cast<DfgVertexUnary*>(vtxp)); }
void DfgVisitor::visit(DfgUnresolved* vtxp) { visit(static_cast<DfgVertexVariadic*>(vtxp)); }
void DfgVisitor::visit(DfgVarArray* vtxp) { visit(static_cast<DfgVertexVar*>(vtxp)); }
void DfgVisitor::visit(DfgVarPacked* vtxp) { visit(static_cast<DfgVertexVar*>(vtxp)); }
void DfgVisitor::visit(DfgVertexAst* vtxp) { visit(static_cast<DfgVertex*>(vtxp)); }
void DfgVisitor::visit(DfgVertexBinary* vtxp) { visit(static_cast<DfgVertex*>(vtxp)); }
void DfgVisitor::visit(DfgVertexNullary* vtxp) { visit(static_cast<DfgVertex*>(vtxp)); }
void DfgVisitor::visit(DfgVertexSplice* vtxp) { visit(static_cast<DfgVertexVariadic*>(vtxp)); }
void DfgVisitor::visit(DfgVertexTernary* vtxp) { visit(static_cast<DfgVertex*>(vtxp)); }
void DfgVisitor::visit(DfgVertexUnary* vtxp) { visit(static_cast<DfgVertex*>(vtxp)); }
void DfgVisitor::visit(DfgVertexVar* vtxp) { visit(static_cast<DfgVertex*>(vtxp)); }
void DfgVisitor::visit(DfgVertexVariadic* vtxp) { visit(static_cast<DfgVertex*>(vtxp)); }
void DfgVisitor::visit(DfgXor* vtxp) { visit(static_cast<DfgVertexBinary*>(vtxp)); }
