// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
void visit(DfgAdd* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstAdd>(vtxp, op1p, op2p);
}
void visit(DfgAnd* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstAnd>(vtxp, op1p, op2p);
}
void visit(DfgArraySel* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstArraySel>(vtxp, op1p, op2p);
}
void visit(DfgConcat* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstConcat>(vtxp, op1p, op2p);
}
void visit(DfgCond* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    AstNodeExpr* const op3p = convertDfgVertexToAstNodeExpr(vtxp->inputp(2));
    m_resultp = makeNode<AstCond>(vtxp, op1p, op2p, op3p);
}
void visit(DfgCountOnes* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstCountOnes>(vtxp, op1p);
}
void visit(DfgDiv* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstDiv>(vtxp, op1p, op2p);
}
void visit(DfgDivS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstDivS>(vtxp, op1p, op2p);
}
void visit(DfgEq* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstEq>(vtxp, op1p, op2p);
}
void visit(DfgEqCase* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstEqCase>(vtxp, op1p, op2p);
}
void visit(DfgEqWild* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstEqWild>(vtxp, op1p, op2p);
}
void visit(DfgExtend* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstExtend>(vtxp, op1p);
}
void visit(DfgExtendS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstExtendS>(vtxp, op1p);
}
void visit(DfgGt* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstGt>(vtxp, op1p, op2p);
}
void visit(DfgGtS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstGtS>(vtxp, op1p, op2p);
}
void visit(DfgGte* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstGte>(vtxp, op1p, op2p);
}
void visit(DfgGteS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstGteS>(vtxp, op1p, op2p);
}
void visit(DfgLogAnd* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstLogAnd>(vtxp, op1p, op2p);
}
void visit(DfgLogEq* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstLogEq>(vtxp, op1p, op2p);
}
void visit(DfgLogIf* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstLogIf>(vtxp, op1p, op2p);
}
void visit(DfgLogNot* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstLogNot>(vtxp, op1p);
}
void visit(DfgLogOr* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstLogOr>(vtxp, op1p, op2p);
}
void visit(DfgLt* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstLt>(vtxp, op1p, op2p);
}
void visit(DfgLtS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstLtS>(vtxp, op1p, op2p);
}
void visit(DfgLte* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstLte>(vtxp, op1p, op2p);
}
void visit(DfgLteS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstLteS>(vtxp, op1p, op2p);
}
void visit(DfgModDiv* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstModDiv>(vtxp, op1p, op2p);
}
void visit(DfgModDivS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstModDivS>(vtxp, op1p, op2p);
}
void visit(DfgMul* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstMul>(vtxp, op1p, op2p);
}
void visit(DfgMulS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstMulS>(vtxp, op1p, op2p);
}
void visit(DfgNegate* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstNegate>(vtxp, op1p);
}
void visit(DfgNeq* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstNeq>(vtxp, op1p, op2p);
}
void visit(DfgNeqCase* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstNeqCase>(vtxp, op1p, op2p);
}
void visit(DfgNeqWild* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstNeqWild>(vtxp, op1p, op2p);
}
void visit(DfgNot* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstNot>(vtxp, op1p);
}
void visit(DfgOneHot* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstOneHot>(vtxp, op1p);
}
void visit(DfgOneHot0* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstOneHot0>(vtxp, op1p);
}
void visit(DfgOr* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstOr>(vtxp, op1p, op2p);
}
void visit(DfgPow* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstPow>(vtxp, op1p, op2p);
}
void visit(DfgPowSS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstPowSS>(vtxp, op1p, op2p);
}
void visit(DfgPowSU* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstPowSU>(vtxp, op1p, op2p);
}
void visit(DfgPowUS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstPowUS>(vtxp, op1p, op2p);
}
void visit(DfgRedAnd* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstRedAnd>(vtxp, op1p);
}
void visit(DfgRedOr* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstRedOr>(vtxp, op1p);
}
void visit(DfgRedXor* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    m_resultp = makeNode<AstRedXor>(vtxp, op1p);
}
void visit(DfgShiftL* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstShiftL>(vtxp, op1p, op2p);
}
void visit(DfgShiftR* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstShiftR>(vtxp, op1p, op2p);
}
void visit(DfgShiftRS* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstShiftRS>(vtxp, op1p, op2p);
}
void visit(DfgStreamL* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstStreamL>(vtxp, op1p, op2p);
}
void visit(DfgStreamR* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstStreamR>(vtxp, op1p, op2p);
}
void visit(DfgSub* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstSub>(vtxp, op1p, op2p);
}
void visit(DfgXor* vtxp) override {
    AstNodeExpr* const op1p = convertDfgVertexToAstNodeExpr(vtxp->inputp(0));
    AstNodeExpr* const op2p = convertDfgVertexToAstNodeExpr(vtxp->inputp(1));
    m_resultp = makeNode<AstXor>(vtxp, op1p, op2p);
}
