// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
case VDfgType::Add: {
  DfgAdd* const cp = new DfgAdd{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::And: {
  DfgAnd* const cp = new DfgAnd{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::ArraySel: {
  DfgArraySel* const cp = new DfgArraySel{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Concat: {
  DfgConcat* const cp = new DfgConcat{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Cond: {
  DfgCond* const cp = new DfgCond{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::CountOnes: {
  DfgCountOnes* const cp = new DfgCountOnes{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Div: {
  DfgDiv* const cp = new DfgDiv{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::DivS: {
  DfgDivS* const cp = new DfgDivS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Eq: {
  DfgEq* const cp = new DfgEq{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::EqCase: {
  DfgEqCase* const cp = new DfgEqCase{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::EqWild: {
  DfgEqWild* const cp = new DfgEqWild{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Extend: {
  DfgExtend* const cp = new DfgExtend{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::ExtendS: {
  DfgExtendS* const cp = new DfgExtendS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Gt: {
  DfgGt* const cp = new DfgGt{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::GtS: {
  DfgGtS* const cp = new DfgGtS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Gte: {
  DfgGte* const cp = new DfgGte{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::GteS: {
  DfgGteS* const cp = new DfgGteS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::LogAnd: {
  DfgLogAnd* const cp = new DfgLogAnd{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::LogEq: {
  DfgLogEq* const cp = new DfgLogEq{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::LogIf: {
  DfgLogIf* const cp = new DfgLogIf{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::LogNot: {
  DfgLogNot* const cp = new DfgLogNot{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::LogOr: {
  DfgLogOr* const cp = new DfgLogOr{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Lt: {
  DfgLt* const cp = new DfgLt{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::LtS: {
  DfgLtS* const cp = new DfgLtS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Lte: {
  DfgLte* const cp = new DfgLte{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::LteS: {
  DfgLteS* const cp = new DfgLteS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::ModDiv: {
  DfgModDiv* const cp = new DfgModDiv{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::ModDivS: {
  DfgModDivS* const cp = new DfgModDivS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Mul: {
  DfgMul* const cp = new DfgMul{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::MulS: {
  DfgMulS* const cp = new DfgMulS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Negate: {
  DfgNegate* const cp = new DfgNegate{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Neq: {
  DfgNeq* const cp = new DfgNeq{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::NeqCase: {
  DfgNeqCase* const cp = new DfgNeqCase{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::NeqWild: {
  DfgNeqWild* const cp = new DfgNeqWild{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Not: {
  DfgNot* const cp = new DfgNot{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::OneHot: {
  DfgOneHot* const cp = new DfgOneHot{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::OneHot0: {
  DfgOneHot0* const cp = new DfgOneHot0{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Or: {
  DfgOr* const cp = new DfgOr{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Pow: {
  DfgPow* const cp = new DfgPow{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::PowSS: {
  DfgPowSS* const cp = new DfgPowSS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::PowSU: {
  DfgPowSU* const cp = new DfgPowSU{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::PowUS: {
  DfgPowUS* const cp = new DfgPowUS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::RedAnd: {
  DfgRedAnd* const cp = new DfgRedAnd{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::RedOr: {
  DfgRedOr* const cp = new DfgRedOr{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::RedXor: {
  DfgRedXor* const cp = new DfgRedXor{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::ShiftL: {
  DfgShiftL* const cp = new DfgShiftL{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::ShiftR: {
  DfgShiftR* const cp = new DfgShiftR{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::ShiftRS: {
  DfgShiftRS* const cp = new DfgShiftRS{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::StreamL: {
  DfgStreamL* const cp = new DfgStreamL{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::StreamR: {
  DfgStreamR* const cp = new DfgStreamR{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Sub: {
  DfgSub* const cp = new DfgSub{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}
case VDfgType::Xor: {
  DfgXor* const cp = new DfgXor{*clonep, vtx.fileline(), vtx.dtype()};
  vtxp2clonep.emplace(&vtx, cp);
  break;
}

