// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
virtual void visit(DfgAdd*);
virtual void visit(DfgAnd*);
virtual void visit(DfgArraySel*);
virtual void visit(DfgAstRd*);
virtual void visit(DfgCReset*);
virtual void visit(DfgConcat*);
virtual void visit(DfgCond*);
virtual void visit(DfgConst*);
virtual void visit(DfgCountOnes*);
virtual void visit(DfgDiv*);
virtual void visit(DfgDivS*);
virtual void visit(DfgEq*);
virtual void visit(DfgEqCase*);
virtual void visit(DfgEqWild*);
virtual void visit(DfgExtend*);
virtual void visit(DfgExtendS*);
virtual void visit(DfgGt*);
virtual void visit(DfgGtS*);
virtual void visit(DfgGte*);
virtual void visit(DfgGteS*);
virtual void visit(DfgLogAnd*);
virtual void visit(DfgLogEq*);
virtual void visit(DfgLogIf*);
virtual void visit(DfgLogNot*);
virtual void visit(DfgLogOr*);
virtual void visit(DfgLogic*);
virtual void visit(DfgLt*);
virtual void visit(DfgLtS*);
virtual void visit(DfgLte*);
virtual void visit(DfgLteS*);
virtual void visit(DfgMatchMasked*);
virtual void visit(DfgModDiv*);
virtual void visit(DfgModDivS*);
virtual void visit(DfgMul*);
virtual void visit(DfgMulS*);
virtual void visit(DfgMux*);
virtual void visit(DfgNegate*);
virtual void visit(DfgNeq*);
virtual void visit(DfgNeqCase*);
virtual void visit(DfgNeqWild*);
virtual void visit(DfgNot*);
virtual void visit(DfgOneHot*);
virtual void visit(DfgOneHot0*);
virtual void visit(DfgOr*);
virtual void visit(DfgPow*);
virtual void visit(DfgPowSS*);
virtual void visit(DfgPowSU*);
virtual void visit(DfgPowUS*);
virtual void visit(DfgRedAnd*);
virtual void visit(DfgRedOr*);
virtual void visit(DfgRedXor*);
virtual void visit(DfgRep*);
virtual void visit(DfgSel*);
virtual void visit(DfgShiftL*);
virtual void visit(DfgShiftR*);
virtual void visit(DfgShiftRS*);
virtual void visit(DfgSpliceArray*);
virtual void visit(DfgSplicePacked*);
virtual void visit(DfgStreamL*);
virtual void visit(DfgStreamR*);
virtual void visit(DfgSub*);
virtual void visit(DfgUnitArray*);
virtual void visit(DfgUnresolved*);
virtual void visit(DfgVarArray*);
virtual void visit(DfgVarPacked*);
virtual void visit(DfgVertexAst*);
virtual void visit(DfgVertexBinary*);
virtual void visit(DfgVertexNullary*);
virtual void visit(DfgVertexSplice*);
virtual void visit(DfgVertexTernary*);
virtual void visit(DfgVertexUnary*);
virtual void visit(DfgVertexVar*);
virtual void visit(DfgVertexVariadic*);
virtual void visit(DfgXor*);
