// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
class DfgAdd final : public DfgVertexBinary {
public:
    DfgAdd(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgAdd;
};
class DfgAnd final : public DfgVertexBinary {
public:
    DfgAnd(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgAnd;
};
class DfgArraySel final : public DfgVertexBinary {
public:
    DfgArraySel(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgArraySel;
};
class DfgConcat final : public DfgVertexBinary {
public:
    DfgConcat(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgConcat;
};
class DfgCond final : public DfgVertexTernary {
public:
    DfgCond(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexTernary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgCond;
};
class DfgCountOnes final : public DfgVertexUnary {
public:
    DfgCountOnes(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgCountOnes;
};
class DfgDiv final : public DfgVertexBinary {
public:
    DfgDiv(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgDiv;
};
class DfgDivS final : public DfgVertexBinary {
public:
    DfgDivS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgDivS;
};
class DfgEq final : public DfgVertexBinary {
public:
    DfgEq(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgEq;
};
class DfgEqCase final : public DfgVertexBinary {
public:
    DfgEqCase(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgEqCase;
};
class DfgEqWild final : public DfgVertexBinary {
public:
    DfgEqWild(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgEqWild;
};
class DfgExtend final : public DfgVertexUnary {
public:
    DfgExtend(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgExtend;
};
class DfgExtendS final : public DfgVertexUnary {
public:
    DfgExtendS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgExtendS;
};
class DfgGt final : public DfgVertexBinary {
public:
    DfgGt(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgGt;
};
class DfgGtS final : public DfgVertexBinary {
public:
    DfgGtS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgGtS;
};
class DfgGte final : public DfgVertexBinary {
public:
    DfgGte(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgGte;
};
class DfgGteS final : public DfgVertexBinary {
public:
    DfgGteS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgGteS;
};
class DfgLogAnd final : public DfgVertexBinary {
public:
    DfgLogAnd(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLogAnd;
};
class DfgLogEq final : public DfgVertexBinary {
public:
    DfgLogEq(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLogEq;
};
class DfgLogIf final : public DfgVertexBinary {
public:
    DfgLogIf(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLogIf;
};
class DfgLogNot final : public DfgVertexUnary {
public:
    DfgLogNot(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLogNot;
};
class DfgLogOr final : public DfgVertexBinary {
public:
    DfgLogOr(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLogOr;
};
class DfgLt final : public DfgVertexBinary {
public:
    DfgLt(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLt;
};
class DfgLtS final : public DfgVertexBinary {
public:
    DfgLtS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLtS;
};
class DfgLte final : public DfgVertexBinary {
public:
    DfgLte(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLte;
};
class DfgLteS final : public DfgVertexBinary {
public:
    DfgLteS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgLteS;
};
class DfgModDiv final : public DfgVertexBinary {
public:
    DfgModDiv(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgModDiv;
};
class DfgModDivS final : public DfgVertexBinary {
public:
    DfgModDivS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgModDivS;
};
class DfgMul final : public DfgVertexBinary {
public:
    DfgMul(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgMul;
};
class DfgMulS final : public DfgVertexBinary {
public:
    DfgMulS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgMulS;
};
class DfgNegate final : public DfgVertexUnary {
public:
    DfgNegate(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgNegate;
};
class DfgNeq final : public DfgVertexBinary {
public:
    DfgNeq(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgNeq;
};
class DfgNeqCase final : public DfgVertexBinary {
public:
    DfgNeqCase(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgNeqCase;
};
class DfgNeqWild final : public DfgVertexBinary {
public:
    DfgNeqWild(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgNeqWild;
};
class DfgNot final : public DfgVertexUnary {
public:
    DfgNot(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgNot;
};
class DfgOneHot final : public DfgVertexUnary {
public:
    DfgOneHot(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgOneHot;
};
class DfgOneHot0 final : public DfgVertexUnary {
public:
    DfgOneHot0(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgOneHot0;
};
class DfgOr final : public DfgVertexBinary {
public:
    DfgOr(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgOr;
};
class DfgPow final : public DfgVertexBinary {
public:
    DfgPow(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgPow;
};
class DfgPowSS final : public DfgVertexBinary {
public:
    DfgPowSS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgPowSS;
};
class DfgPowSU final : public DfgVertexBinary {
public:
    DfgPowSU(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgPowSU;
};
class DfgPowUS final : public DfgVertexBinary {
public:
    DfgPowUS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgPowUS;
};
class DfgRedAnd final : public DfgVertexUnary {
public:
    DfgRedAnd(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgRedAnd;
};
class DfgRedOr final : public DfgVertexUnary {
public:
    DfgRedOr(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgRedOr;
};
class DfgRedXor final : public DfgVertexUnary {
public:
    DfgRedXor(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexUnary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgRedXor;
};
class DfgShiftL final : public DfgVertexBinary {
public:
    DfgShiftL(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgShiftL;
};
class DfgShiftR final : public DfgVertexBinary {
public:
    DfgShiftR(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgShiftR;
};
class DfgShiftRS final : public DfgVertexBinary {
public:
    DfgShiftRS(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgShiftRS;
};
class DfgStreamL final : public DfgVertexBinary {
public:
    DfgStreamL(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgStreamL;
};
class DfgStreamR final : public DfgVertexBinary {
public:
    DfgStreamR(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgStreamR;
};
class DfgSub final : public DfgVertexBinary {
public:
    DfgSub(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgSub;
};
class DfgXor final : public DfgVertexBinary {
public:
    DfgXor(DfgGraph& dfg, FileLine* flp, const DfgDataType& dtype)
        : DfgVertexBinary{dfg, dfgType(), flp, dtype} {}
    ASTGEN_MEMBERS_DfgXor;
};

