// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
virtual void visit(AstAbortOn*);
virtual void visit(AstAcosD*);
virtual void visit(AstAcoshD*);
virtual void visit(AstActive*);
virtual void visit(AstAdd*);
virtual void visit(AstAddD*);
virtual void visit(AstAddrOfCFunc*);
virtual void visit(AstAlias*);
virtual void visit(AstAliasScope*);
virtual void visit(AstAlways*);
virtual void visit(AstAlwaysObserved*);
virtual void visit(AstAlwaysPost*);
virtual void visit(AstAlwaysPostponed*);
virtual void visit(AstAlwaysPre*);
virtual void visit(AstAlwaysReactive*);
virtual void visit(AstAnd*);
virtual void visit(AstArg*);
virtual void visit(AstArraySel*);
virtual void visit(AstAsinD*);
virtual void visit(AstAsinhD*);
virtual void visit(AstAssert*);
virtual void visit(AstAssertCtl*);
virtual void visit(AstAssertIntrinsic*);
virtual void visit(AstAssign*);
virtual void visit(AstAssignCompound*);
virtual void visit(AstAssignCont*);
virtual void visit(AstAssignDly*);
virtual void visit(AstAssignForce*);
virtual void visit(AstAssignW*);
virtual void visit(AstAssocArrayDType*);
virtual void visit(AstAssocSel*);
virtual void visit(AstAtan2D*);
virtual void visit(AstAtanD*);
virtual void visit(AstAtanhD*);
virtual void visit(AstAtoN*);
virtual void visit(AstAttrOf*);
virtual void visit(AstBasicDType*);
virtual void visit(AstBegin*);
virtual void visit(AstBind*);
virtual void visit(AstBitsToRealD*);
virtual void visit(AstBracketArrayDType*);
virtual void visit(AstBracketRange*);
virtual void visit(AstBreak*);
virtual void visit(AstBufIf1*);
virtual void visit(AstCAwait*);
virtual void visit(AstCCall*);
virtual void visit(AstCCast*);
virtual void visit(AstCDType*);
virtual void visit(AstCExpr*);
virtual void visit(AstCExprUser*);
virtual void visit(AstCFile*);
virtual void visit(AstCFunc*);
virtual void visit(AstCLocalScope*);
virtual void visit(AstCLog2*);
virtual void visit(AstCMethodCall*);
virtual void visit(AstCMethodHard*);
virtual void visit(AstCNew*);
virtual void visit(AstCReset*);
virtual void visit(AstCReturn*);
virtual void visit(AstCStmt*);
virtual void visit(AstCStmtUser*);
virtual void visit(AstCUse*);
virtual void visit(AstCase*);
virtual void visit(AstCaseItem*);
virtual void visit(AstCast*);
virtual void visit(AstCastDynamic*);
virtual void visit(AstCastParse*);
virtual void visit(AstCastSize*);
virtual void visit(AstCastWrap*);
virtual void visit(AstCeilD*);
virtual void visit(AstCell*);
virtual void visit(AstCellArrayRef*);
virtual void visit(AstCellInline*);
virtual void visit(AstCellInlineScope*);
virtual void visit(AstCellRef*);
virtual void visit(AstCgOptionAssign*);
virtual void visit(AstClass*);
virtual void visit(AstClassExtends*);
virtual void visit(AstClassOrPackageRef*);
virtual void visit(AstClassPackage*);
virtual void visit(AstClassRefDType*);
virtual void visit(AstClocking*);
virtual void visit(AstClockingItem*);
virtual void visit(AstComment*);
virtual void visit(AstCompareNN*);
virtual void visit(AstConcat*);
virtual void visit(AstConcatN*);
virtual void visit(AstCond*);
virtual void visit(AstConfig*);
virtual void visit(AstConfigCell*);
virtual void visit(AstConfigRule*);
virtual void visit(AstConfigUse*);
virtual void visit(AstConsAssoc*);
virtual void visit(AstConsDynArray*);
virtual void visit(AstConsPackMember*);
virtual void visit(AstConsPackUOrStruct*);
virtual void visit(AstConsQueue*);
virtual void visit(AstConsWildcard*);
virtual void visit(AstConst*);
virtual void visit(AstConstDType*);
virtual void visit(AstConstPool*);
virtual void visit(AstConstraint*);
virtual void visit(AstConstraintBefore*);
virtual void visit(AstConstraintExpr*);
virtual void visit(AstConstraintForeach*);
virtual void visit(AstConstraintIf*);
virtual void visit(AstConstraintRef*);
virtual void visit(AstConstraintRefDType*);
virtual void visit(AstConstraintUnique*);
virtual void visit(AstContinue*);
virtual void visit(AstCosD*);
virtual void visit(AstCoshD*);
virtual void visit(AstCountBits*);
virtual void visit(AstCountOnes*);
virtual void visit(AstCover*);
virtual void visit(AstCoverBin*);
virtual void visit(AstCoverCross*);
virtual void visit(AstCoverInc*);
virtual void visit(AstCoverOption*);
virtual void visit(AstCoverOtherDecl*);
virtual void visit(AstCoverToggle*);
virtual void visit(AstCoverToggleDecl*);
virtual void visit(AstCoverTransItem*);
virtual void visit(AstCoverTransSet*);
virtual void visit(AstCovergroup*);
virtual void visit(AstCoverpoint*);
virtual void visit(AstCoverpointRef*);
virtual void visit(AstCvtArrayToArray*);
virtual void visit(AstCvtArrayToPacked*);
virtual void visit(AstCvtPackString*);
virtual void visit(AstCvtPackedToArray*);
virtual void visit(AstCvtUnpackedToQueue*);
virtual void visit(AstDeassign*);
virtual void visit(AstDefImplicitDType*);
virtual void visit(AstDefParam*);
virtual void visit(AstDefaultDisable*);
virtual void visit(AstDelay*);
virtual void visit(AstDisable*);
virtual void visit(AstDisableFork*);
virtual void visit(AstDisplay*);
virtual void visit(AstDist*);
virtual void visit(AstDistChiSquare*);
virtual void visit(AstDistErlang*);
virtual void visit(AstDistExponential*);
virtual void visit(AstDistItem*);
virtual void visit(AstDistNormal*);
virtual void visit(AstDistPoisson*);
virtual void visit(AstDistT*);
virtual void visit(AstDistUniform*);
virtual void visit(AstDiv*);
virtual void visit(AstDivD*);
virtual void visit(AstDivS*);
virtual void visit(AstDot*);
virtual void visit(AstDpiExport*);
virtual void visit(AstDumpCtl*);
virtual void visit(AstDynArrayDType*);
virtual void visit(AstElabDisplay*);
virtual void visit(AstEmpty*);
virtual void visit(AstEmptyQueue*);
virtual void visit(AstEmptyQueueDType*);
virtual void visit(AstEnumDType*);
virtual void visit(AstEnumItem*);
virtual void visit(AstEnumItemRef*);
virtual void visit(AstEq*);
virtual void visit(AstEqCase*);
virtual void visit(AstEqD*);
virtual void visit(AstEqN*);
virtual void visit(AstEqT*);
virtual void visit(AstEqWild*);
virtual void visit(AstEventControl*);
virtual void visit(AstExecGraph*);
virtual void visit(AstExpD*);
virtual void visit(AstExprStmt*);
virtual void visit(AstExtend*);
virtual void visit(AstExtendS*);
virtual void visit(AstFClose*);
virtual void visit(AstFEof*);
virtual void visit(AstFError*);
virtual void visit(AstFFlush*);
virtual void visit(AstFGetC*);
virtual void visit(AstFGetS*);
virtual void visit(AstFOpen*);
virtual void visit(AstFOpenMcd*);
virtual void visit(AstFRead*);
virtual void visit(AstFRewind*);
virtual void visit(AstFScanF*);
virtual void visit(AstFSeek*);
virtual void visit(AstFTell*);
virtual void visit(AstFUngetC*);
virtual void visit(AstFalling*);
virtual void visit(AstFell*);
virtual void visit(AstFinal*);
virtual void visit(AstFinish*);
virtual void visit(AstFinishFork*);
virtual void visit(AstFireEvent*);
virtual void visit(AstFloorD*);
virtual void visit(AstForeach*);
virtual void visit(AstForeachHeader*);
virtual void visit(AstFork*);
virtual void visit(AstFunc*);
virtual void visit(AstFuncRef*);
virtual void visit(AstFuture*);
virtual void visit(AstGatePin*);
virtual void visit(AstGenBlock*);
virtual void visit(AstGenCase*);
virtual void visit(AstGenCaseItem*);
virtual void visit(AstGenFor*);
virtual void visit(AstGenIf*);
virtual void visit(AstGetInitialRandomSeed*);
virtual void visit(AstGetcN*);
virtual void visit(AstGetcRefN*);
virtual void visit(AstGt*);
virtual void visit(AstGtD*);
virtual void visit(AstGtN*);
virtual void visit(AstGtS*);
virtual void visit(AstGte*);
virtual void visit(AstGteD*);
virtual void visit(AstGteN*);
virtual void visit(AstGteS*);
virtual void visit(AstHypotD*);
virtual void visit(AstISToRD*);
virtual void visit(AstIToRD*);
virtual void visit(AstIf*);
virtual void visit(AstIface*);
virtual void visit(AstIfaceGenericDType*);
virtual void visit(AstIfaceRefDType*);
virtual void visit(AstImplication*);
virtual void visit(AstImplicit*);
virtual void visit(AstInferredDisable*);
virtual void visit(AstInitArray*);
virtual void visit(AstInitItem*);
virtual void visit(AstInitial*);
virtual void visit(AstInitialAutomatic*);
virtual void visit(AstInitialAutomaticStmt*);
virtual void visit(AstInitialStatic*);
virtual void visit(AstInitialStaticStmt*);
virtual void visit(AstInside*);
virtual void visit(AstInsideRange*);
virtual void visit(AstIntfRef*);
virtual void visit(AstIsUnbounded*);
virtual void visit(AstIsUnknown*);
virtual void visit(AstJumpBlock*);
virtual void visit(AstJumpGo*);
virtual void visit(AstLambdaArgRef*);
virtual void visit(AstLenN*);
virtual void visit(AstLet*);
virtual void visit(AstLibrary*);
virtual void visit(AstLog10D*);
virtual void visit(AstLogAnd*);
virtual void visit(AstLogD*);
virtual void visit(AstLogEq*);
virtual void visit(AstLogIf*);
virtual void visit(AstLogNot*);
virtual void visit(AstLogOr*);
virtual void visit(AstLoop*);
virtual void visit(AstLoopTest*);
virtual void visit(AstLt*);
virtual void visit(AstLtD*);
virtual void visit(AstLtN*);
virtual void visit(AstLtS*);
virtual void visit(AstLte*);
virtual void visit(AstLteD*);
virtual void visit(AstLteN*);
virtual void visit(AstLteS*);
virtual void visit(AstMatchMasked*);
virtual void visit(AstMatches*);
virtual void visit(AstMemberDType*);
virtual void visit(AstMemberSel*);
virtual void visit(AstMethodCall*);
virtual void visit(AstModDiv*);
virtual void visit(AstModDivS*);
virtual void visit(AstModport*);
virtual void visit(AstModportClockingRef*);
virtual void visit(AstModportFTaskRef*);
virtual void visit(AstModportVarRef*);
virtual void visit(AstModule*);
virtual void visit(AstMonitorOff*);
virtual void visit(AstMostSetBitP1*);
virtual void visit(AstMul*);
virtual void visit(AstMulD*);
virtual void visit(AstMulS*);
virtual void visit(AstNBACommitQueueDType*);
virtual void visit(AstNToI*);
virtual void visit(AstNegate*);
virtual void visit(AstNegateD*);
virtual void visit(AstNeq*);
virtual void visit(AstNeqCase*);
virtual void visit(AstNeqD*);
virtual void visit(AstNeqN*);
virtual void visit(AstNeqT*);
virtual void visit(AstNeqWild*);
virtual void visit(AstNetlist*);
virtual void visit(AstNew*);
virtual void visit(AstNewCopy*);
virtual void visit(AstNewDynamic*);
virtual void visit(AstNodeArrayDType*);
virtual void visit(AstNodeAssign*);
virtual void visit(AstNodeBiCom*);
virtual void visit(AstNodeBiComAsv*);
virtual void visit(AstNodeBiop*);
virtual void visit(AstNodeBlock*);
virtual void visit(AstNodeCCall*);
virtual void visit(AstNodeCoverDecl*);
virtual void visit(AstNodeCoverOrAssert*);
virtual void visit(AstNodeDType*);
virtual void visit(AstNodeDistBiop*);
virtual void visit(AstNodeDistTriop*);
virtual void visit(AstNodeExpr*);
virtual void visit(AstNodeFTask*);
virtual void visit(AstNodeFTaskRef*);
virtual void visit(AstNodeFile*);
virtual void visit(AstNodeForeach*);
virtual void visit(AstNodeFuncCovItem*);
virtual void visit(AstNodeGen*);
virtual void visit(AstNodeIf*);
virtual void visit(AstNodeModule*);
virtual void visit(AstNodePreSel*);
virtual void visit(AstNodeProcedure*);
virtual void visit(AstNodeQuadop*);
virtual void visit(AstNodeRange*);
virtual void visit(AstNodeReadWriteMem*);
virtual void visit(AstNodeSel*);
virtual void visit(AstNodeStmt*);
virtual void visit(AstNodeStream*);
virtual void visit(AstNodeSystemBiopD*);
virtual void visit(AstNodeSystemUniopD*);
virtual void visit(AstNodeTermop*);
virtual void visit(AstNodeTriop*);
virtual void visit(AstNodeUOrStructDType*);
virtual void visit(AstNodeUniop*);
virtual void visit(AstNodeVarRef*);
virtual void visit(AstNot*);
virtual void visit(AstNotFoundModule*);
virtual void visit(AstNullCheck*);
virtual void visit(AstOneHot*);
virtual void visit(AstOneHot0*);
virtual void visit(AstOr*);
virtual void visit(AstPExpr*);
virtual void visit(AstPExprClause*);
virtual void visit(AstPackArrayDType*);
virtual void visit(AstPackage*);
virtual void visit(AstPackageExport*);
virtual void visit(AstPackageExportStarStar*);
virtual void visit(AstPackageImport*);
virtual void visit(AstParamTypeDType*);
virtual void visit(AstParseHolder*);
virtual void visit(AstParseRef*);
virtual void visit(AstParseTypeDType*);
virtual void visit(AstPast*);
virtual void visit(AstPatMember*);
virtual void visit(AstPattern*);
virtual void visit(AstPatternStar*);
virtual void visit(AstPatternVar*);
virtual void visit(AstPin*);
virtual void visit(AstPort*);
virtual void visit(AstPostDec*);
virtual void visit(AstPostInc*);
virtual void visit(AstPow*);
virtual void visit(AstPowD*);
virtual void visit(AstPowSS*);
virtual void visit(AstPowSU*);
virtual void visit(AstPowUS*);
virtual void visit(AstPragma*);
virtual void visit(AstPreDec*);
virtual void visit(AstPreInc*);
virtual void visit(AstPrimitive*);
virtual void visit(AstPrintTimeScale*);
virtual void visit(AstPropAlways*);
virtual void visit(AstPropSpec*);
virtual void visit(AstProperty*);
virtual void visit(AstPull*);
virtual void visit(AstPutcN*);
virtual void visit(AstQueueDType*);
virtual void visit(AstRSBreak*);
virtual void visit(AstRSProd*);
virtual void visit(AstRSProdItem*);
virtual void visit(AstRSProdList*);
virtual void visit(AstRSReturn*);
virtual void visit(AstRSRule*);
virtual void visit(AstRToIRoundS*);
virtual void visit(AstRToIS*);
virtual void visit(AstRand*);
virtual void visit(AstRandCase*);
virtual void visit(AstRandRNG*);
virtual void visit(AstRandSequence*);
virtual void visit(AstRange*);
virtual void visit(AstReadMem*);
virtual void visit(AstRealToBits*);
virtual void visit(AstRedAnd*);
virtual void visit(AstRedOr*);
virtual void visit(AstRedXor*);
virtual void visit(AstRefDType*);
virtual void visit(AstRelease*);
virtual void visit(AstRepeat*);
virtual void visit(AstReplicate*);
virtual void visit(AstReplicateN*);
virtual void visit(AstRequireDType*);
virtual void visit(AstResizeLValue*);
virtual void visit(AstRestrict*);
virtual void visit(AstReturn*);
virtual void visit(AstRising*);
virtual void visit(AstRose*);
virtual void visit(AstSAnd*);
virtual void visit(AstSClocked*);
virtual void visit(AstSConsRep*);
virtual void visit(AstSEventually*);
virtual void visit(AstSExpr*);
virtual void visit(AstSFormat*);
virtual void visit(AstSFormatArg*);
virtual void visit(AstSFormatF*);
virtual void visit(AstSGotoRep*);
virtual void visit(AstSIntersect*);
virtual void visit(AstSNonConsRep*);
virtual void visit(AstSOr*);
virtual void visit(AstSScanF*);
virtual void visit(AstSThroughout*);
virtual void visit(AstSWithin*);
virtual void visit(AstSampleQueueDType*);
virtual void visit(AstSampled*);
virtual void visit(AstScope*);
virtual void visit(AstScopeName*);
virtual void visit(AstSel*);
virtual void visit(AstSelBit*);
virtual void visit(AstSelExtract*);
virtual void visit(AstSelMinus*);
virtual void visit(AstSelPlus*);
virtual void visit(AstSenItem*);
virtual void visit(AstSenTree*);
virtual void visit(AstSequence*);
virtual void visit(AstSetAssoc*);
virtual void visit(AstSetWildcard*);
virtual void visit(AstSetuphold*);
virtual void visit(AstShiftL*);
virtual void visit(AstShiftLOvr*);
virtual void visit(AstShiftR*);
virtual void visit(AstShiftROvr*);
virtual void visit(AstShiftRS*);
virtual void visit(AstShiftRSOvr*);
virtual void visit(AstSigned*);
virtual void visit(AstSinD*);
virtual void visit(AstSinhD*);
virtual void visit(AstSliceSel*);
virtual void visit(AstSplitPlaceholder*);
virtual void visit(AstSqrtD*);
virtual void visit(AstStable*);
virtual void visit(AstStackTraceF*);
virtual void visit(AstStackTraceT*);
virtual void visit(AstSteady*);
virtual void visit(AstStmtExpr*);
virtual void visit(AstStmtPragma*);
virtual void visit(AstStop*);
virtual void visit(AstStreamDType*);
virtual void visit(AstStreamL*);
virtual void visit(AstStreamR*);
virtual void visit(AstStrengthSpec*);
virtual void visit(AstStructDType*);
virtual void visit(AstStructSel*);
virtual void visit(AstSub*);
virtual void visit(AstSubD*);
virtual void visit(AstSubstrN*);
virtual void visit(AstSysIgnore*);
virtual void visit(AstSystemCSection*);
virtual void visit(AstSystemF*);
virtual void visit(AstSystemT*);
virtual void visit(AstTaggedExpr*);
virtual void visit(AstTaggedPattern*);
virtual void visit(AstTanD*);
virtual void visit(AstTanhD*);
virtual void visit(AstTask*);
virtual void visit(AstTaskRef*);
virtual void visit(AstTestPlusArgs*);
virtual void visit(AstText*);
virtual void visit(AstTextBlock*);
virtual void visit(AstThisRef*);
virtual void visit(AstTime*);
virtual void visit(AstTimeD*);
virtual void visit(AstTimeFormat*);
virtual void visit(AstTimeImport*);
virtual void visit(AstTimePrecision*);
virtual void visit(AstTimeUnit*);
virtual void visit(AstToLowerN*);
virtual void visit(AstToStringN*);
virtual void visit(AstToUpperN*);
virtual void visit(AstTopScope*);
virtual void visit(AstTraceDecl*);
virtual void visit(AstTraceInc*);
virtual void visit(AstTracePopPrefix*);
virtual void visit(AstTracePushPrefix*);
virtual void visit(AstTypeTable*);
virtual void visit(AstTypedef*);
virtual void visit(AstTypedefFwd*);
virtual void visit(AstURandomRange*);
virtual void visit(AstUdpTable*);
virtual void visit(AstUdpTableLine*);
virtual void visit(AstUdpTableLineVal*);
virtual void visit(AstUnbounded*);
virtual void visit(AstUnionDType*);
virtual void visit(AstUnlinkedRef*);
virtual void visit(AstUnpackArrayDType*);
virtual void visit(AstUnsigned*);
virtual void visit(AstUnsizedArrayDType*);
virtual void visit(AstUnsizedRange*);
virtual void visit(AstUntil*);
virtual void visit(AstVFile*);
virtual void visit(AstValuePlusArgs*);
virtual void visit(AstVar*);
virtual void visit(AstVarRef*);
virtual void visit(AstVarScope*);
virtual void visit(AstVarXRef*);
virtual void visit(AstVoidDType*);
virtual void visit(AstWait*);
virtual void visit(AstWaitFork*);
virtual void visit(AstWildcardArrayDType*);
virtual void visit(AstWildcardRange*);
virtual void visit(AstWildcardSel*);
virtual void visit(AstWith*);
virtual void visit(AstWithParse*);
virtual void visit(AstWordSel*);
virtual void visit(AstWriteMem*);
virtual void visit(AstXor*);
