// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
void VNVisitorConst::visit(AstAbortOn* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstAcosD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstAcoshD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstActive* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstAdd* nodep) { visit(static_cast<AstNodeBiComAsv*>(nodep)); }
void VNVisitorConst::visit(AstAddD* nodep) { visit(static_cast<AstNodeBiComAsv*>(nodep)); }
void VNVisitorConst::visit(AstAddrOfCFunc* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstAlias* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstAliasScope* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstAlways* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstAlwaysObserved* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstAlwaysPost* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstAlwaysPostponed* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstAlwaysPre* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstAlwaysReactive* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstAnd* nodep) { visit(static_cast<AstNodeBiComAsv*>(nodep)); }
void VNVisitorConst::visit(AstArg* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstArraySel* nodep) { visit(static_cast<AstNodeSel*>(nodep)); }
void VNVisitorConst::visit(AstAsinD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstAsinhD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstAssert* nodep) { visit(static_cast<AstNodeCoverOrAssert*>(nodep)); }
void VNVisitorConst::visit(AstAssertCtl* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstAssertIntrinsic* nodep) { visit(static_cast<AstNodeCoverOrAssert*>(nodep)); }
void VNVisitorConst::visit(AstAssign* nodep) { visit(static_cast<AstNodeAssign*>(nodep)); }
void VNVisitorConst::visit(AstAssignCompound* nodep) { visit(static_cast<AstNodeAssign*>(nodep)); }
void VNVisitorConst::visit(AstAssignCont* nodep) { visit(static_cast<AstNodeAssign*>(nodep)); }
void VNVisitorConst::visit(AstAssignDly* nodep) { visit(static_cast<AstNodeAssign*>(nodep)); }
void VNVisitorConst::visit(AstAssignForce* nodep) { visit(static_cast<AstNodeAssign*>(nodep)); }
void VNVisitorConst::visit(AstAssignW* nodep) { visit(static_cast<AstNodeAssign*>(nodep)); }
void VNVisitorConst::visit(AstAssocArrayDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstAssocSel* nodep) { visit(static_cast<AstNodeSel*>(nodep)); }
void VNVisitorConst::visit(AstAtan2D* nodep) { visit(static_cast<AstNodeSystemBiopD*>(nodep)); }
void VNVisitorConst::visit(AstAtanD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstAtanhD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstAtoN* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstAttrOf* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstBasicDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstBegin* nodep) { visit(static_cast<AstNodeBlock*>(nodep)); }
void VNVisitorConst::visit(AstBind* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstBitsToRealD* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstBracketArrayDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstBracketRange* nodep) { visit(static_cast<AstNodeRange*>(nodep)); }
void VNVisitorConst::visit(AstBreak* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstBufIf1* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstCAwait* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCCall* nodep) { visit(static_cast<AstNodeCCall*>(nodep)); }
void VNVisitorConst::visit(AstCCast* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstCDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstCExpr* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCExprUser* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCFile* nodep) { visit(static_cast<AstNodeFile*>(nodep)); }
void VNVisitorConst::visit(AstCFunc* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCLocalScope* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCLog2* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstCMethodCall* nodep) { visit(static_cast<AstNodeCCall*>(nodep)); }
void VNVisitorConst::visit(AstCMethodHard* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCNew* nodep) { visit(static_cast<AstNodeCCall*>(nodep)); }
void VNVisitorConst::visit(AstCReset* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCReturn* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCStmt* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCStmtUser* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCUse* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCase* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCaseItem* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCast* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCastDynamic* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstCastParse* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCastSize* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCastWrap* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstCeilD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstCell* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCellArrayRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCellInline* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCellInlineScope* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCellRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCgOptionAssign* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstClass* nodep) { visit(static_cast<AstNodeModule*>(nodep)); }
void VNVisitorConst::visit(AstClassExtends* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstClassOrPackageRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstClassPackage* nodep) { visit(static_cast<AstNodeModule*>(nodep)); }
void VNVisitorConst::visit(AstClassRefDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstClocking* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstClockingItem* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstComment* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCompareNN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstConcat* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstConcatN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstCond* nodep) { visit(static_cast<AstNodeTriop*>(nodep)); }
void VNVisitorConst::visit(AstConfig* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstConfigCell* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstConfigRule* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstConfigUse* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstConsAssoc* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstConsDynArray* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstConsPackMember* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstConsPackUOrStruct* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstConsQueue* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstConsWildcard* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstConst* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstConstDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstConstPool* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstConstraint* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstConstraintBefore* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstConstraintExpr* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstConstraintForeach* nodep) { visit(static_cast<AstNodeForeach*>(nodep)); }
void VNVisitorConst::visit(AstConstraintIf* nodep) { visit(static_cast<AstNodeIf*>(nodep)); }
void VNVisitorConst::visit(AstConstraintRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstConstraintRefDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstConstraintUnique* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstContinue* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCosD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstCoshD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstCountBits* nodep) { visit(static_cast<AstNodeQuadop*>(nodep)); }
void VNVisitorConst::visit(AstCountOnes* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstCover* nodep) { visit(static_cast<AstNodeCoverOrAssert*>(nodep)); }
void VNVisitorConst::visit(AstCoverBin* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCoverCross* nodep) { visit(static_cast<AstNodeFuncCovItem*>(nodep)); }
void VNVisitorConst::visit(AstCoverInc* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCoverOption* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCoverOtherDecl* nodep) { visit(static_cast<AstNodeCoverDecl*>(nodep)); }
void VNVisitorConst::visit(AstCoverToggle* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstCoverToggleDecl* nodep) { visit(static_cast<AstNodeCoverDecl*>(nodep)); }
void VNVisitorConst::visit(AstCoverTransItem* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCoverTransSet* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCovergroup* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCoverpoint* nodep) { visit(static_cast<AstNodeFuncCovItem*>(nodep)); }
void VNVisitorConst::visit(AstCoverpointRef* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstCvtArrayToArray* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCvtArrayToPacked* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCvtPackString* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstCvtPackedToArray* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstCvtUnpackedToQueue* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstDeassign* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstDefImplicitDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstDefParam* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstDefaultDisable* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstDelay* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstDisable* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstDisableFork* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstDisplay* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstDist* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstDistChiSquare* nodep) { visit(static_cast<AstNodeDistBiop*>(nodep)); }
void VNVisitorConst::visit(AstDistErlang* nodep) { visit(static_cast<AstNodeDistTriop*>(nodep)); }
void VNVisitorConst::visit(AstDistExponential* nodep) { visit(static_cast<AstNodeDistBiop*>(nodep)); }
void VNVisitorConst::visit(AstDistItem* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstDistNormal* nodep) { visit(static_cast<AstNodeDistTriop*>(nodep)); }
void VNVisitorConst::visit(AstDistPoisson* nodep) { visit(static_cast<AstNodeDistBiop*>(nodep)); }
void VNVisitorConst::visit(AstDistT* nodep) { visit(static_cast<AstNodeDistBiop*>(nodep)); }
void VNVisitorConst::visit(AstDistUniform* nodep) { visit(static_cast<AstNodeDistTriop*>(nodep)); }
void VNVisitorConst::visit(AstDiv* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstDivD* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstDivS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstDot* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstDpiExport* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstDumpCtl* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstDynArrayDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstElabDisplay* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstEmpty* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstEmptyQueue* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstEmptyQueueDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstEnumDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstEnumItem* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstEnumItemRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstEq* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstEqCase* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstEqD* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstEqN* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstEqT* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstEqWild* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstEventControl* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstExecGraph* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstExpD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstExprStmt* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstExtend* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstExtendS* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstFClose* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstFEof* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstFError* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFFlush* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstFGetC* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstFGetS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstFOpen* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFOpenMcd* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFRead* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFRewind* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFScanF* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFSeek* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFTell* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFUngetC* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstFalling* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFell* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstFinal* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstFinish* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstFinishFork* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstFireEvent* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstFloorD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstForeach* nodep) { visit(static_cast<AstNodeForeach*>(nodep)); }
void VNVisitorConst::visit(AstForeachHeader* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstFork* nodep) { visit(static_cast<AstNodeBlock*>(nodep)); }
void VNVisitorConst::visit(AstFunc* nodep) { visit(static_cast<AstNodeFTask*>(nodep)); }
void VNVisitorConst::visit(AstFuncRef* nodep) { visit(static_cast<AstNodeFTaskRef*>(nodep)); }
void VNVisitorConst::visit(AstFuture* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstGatePin* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstGenBlock* nodep) { visit(static_cast<AstNodeGen*>(nodep)); }
void VNVisitorConst::visit(AstGenCase* nodep) { visit(static_cast<AstNodeGen*>(nodep)); }
void VNVisitorConst::visit(AstGenCaseItem* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstGenFor* nodep) { visit(static_cast<AstNodeGen*>(nodep)); }
void VNVisitorConst::visit(AstGenIf* nodep) { visit(static_cast<AstNodeGen*>(nodep)); }
void VNVisitorConst::visit(AstGetInitialRandomSeed* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstGetcN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGetcRefN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGt* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGtD* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGtN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGtS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGte* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGteD* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGteN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstGteS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstHypotD* nodep) { visit(static_cast<AstNodeSystemBiopD*>(nodep)); }
void VNVisitorConst::visit(AstISToRD* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstIToRD* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstIf* nodep) { visit(static_cast<AstNodeIf*>(nodep)); }
void VNVisitorConst::visit(AstIface* nodep) { visit(static_cast<AstNodeModule*>(nodep)); }
void VNVisitorConst::visit(AstIfaceGenericDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstIfaceRefDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstImplication* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstImplicit* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstInferredDisable* nodep) { visit(static_cast<AstNodeTermop*>(nodep)); }
void VNVisitorConst::visit(AstInitArray* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstInitItem* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstInitial* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstInitialAutomatic* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstInitialAutomaticStmt* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstInitialStatic* nodep) { visit(static_cast<AstNodeProcedure*>(nodep)); }
void VNVisitorConst::visit(AstInitialStaticStmt* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstInside* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstInsideRange* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstIntfRef* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstIsUnbounded* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstIsUnknown* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstJumpBlock* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstJumpGo* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstLambdaArgRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstLenN* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstLet* nodep) { visit(static_cast<AstNodeFTask*>(nodep)); }
void VNVisitorConst::visit(AstLibrary* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstLog10D* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstLogAnd* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLogD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstLogEq* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstLogIf* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLogNot* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstLogOr* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLoop* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstLoopTest* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstLt* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLtD* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLtN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLtS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLte* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLteD* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLteN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstLteS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstMatchMasked* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstMatches* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstMemberDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstMemberSel* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstMethodCall* nodep) { visit(static_cast<AstNodeFTaskRef*>(nodep)); }
void VNVisitorConst::visit(AstModDiv* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstModDivS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstModport* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstModportClockingRef* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstModportFTaskRef* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstModportVarRef* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstModule* nodep) { visit(static_cast<AstNodeModule*>(nodep)); }
void VNVisitorConst::visit(AstMonitorOff* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstMostSetBitP1* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstMul* nodep) { visit(static_cast<AstNodeBiComAsv*>(nodep)); }
void VNVisitorConst::visit(AstMulD* nodep) { visit(static_cast<AstNodeBiComAsv*>(nodep)); }
void VNVisitorConst::visit(AstMulS* nodep) { visit(static_cast<AstNodeBiComAsv*>(nodep)); }
void VNVisitorConst::visit(AstNBACommitQueueDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstNToI* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstNegate* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstNegateD* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstNeq* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstNeqCase* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstNeqD* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstNeqN* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstNeqT* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstNeqWild* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstNetlist* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNew* nodep) { visit(static_cast<AstNodeFTaskRef*>(nodep)); }
void VNVisitorConst::visit(AstNewCopy* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNewDynamic* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeArrayDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstNodeAssign* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstNodeBiCom* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstNodeBiComAsv* nodep) { visit(static_cast<AstNodeBiCom*>(nodep)); }
void VNVisitorConst::visit(AstNodeBiop* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeBlock* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstNodeCCall* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeCoverDecl* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeCoverOrAssert* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstNodeDType* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeDistBiop* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstNodeDistTriop* nodep) { visit(static_cast<AstNodeTriop*>(nodep)); }
void VNVisitorConst::visit(AstNodeExpr* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeFTask* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeFTaskRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeFile* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeForeach* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstNodeFuncCovItem* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeGen* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeIf* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstNodeModule* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodePreSel* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeProcedure* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeQuadop* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeRange* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeReadWriteMem* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstNodeSel* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstNodeStmt* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstNodeStream* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstNodeSystemBiopD* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstNodeSystemUniopD* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstNodeTermop* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeTriop* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeUOrStructDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstNodeUniop* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNodeVarRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstNot* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstNotFoundModule* nodep) { visit(static_cast<AstNodeModule*>(nodep)); }
void VNVisitorConst::visit(AstNullCheck* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstOneHot* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstOneHot0* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstOr* nodep) { visit(static_cast<AstNodeBiComAsv*>(nodep)); }
void VNVisitorConst::visit(AstPExpr* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstPExprClause* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstPackArrayDType* nodep) { visit(static_cast<AstNodeArrayDType*>(nodep)); }
void VNVisitorConst::visit(AstPackage* nodep) { visit(static_cast<AstNodeModule*>(nodep)); }
void VNVisitorConst::visit(AstPackageExport* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstPackageExportStarStar* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstPackageImport* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstParamTypeDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstParseHolder* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstParseRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstParseTypeDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstPast* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstPatMember* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstPattern* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstPatternStar* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstPatternVar* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstPin* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstPort* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstPostDec* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstPostInc* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstPow* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstPowD* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstPowSS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstPowSU* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstPowUS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstPragma* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstPreDec* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstPreInc* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstPrimitive* nodep) { visit(static_cast<AstNodeModule*>(nodep)); }
void VNVisitorConst::visit(AstPrintTimeScale* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstPropAlways* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstPropSpec* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstProperty* nodep) { visit(static_cast<AstNodeFTask*>(nodep)); }
void VNVisitorConst::visit(AstPull* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstPutcN* nodep) { visit(static_cast<AstNodeTriop*>(nodep)); }
void VNVisitorConst::visit(AstQueueDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstRSBreak* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRSProd* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRSProdItem* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRSProdList* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRSReturn* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRSRule* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRToIRoundS* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstRToIS* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstRand* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstRandCase* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRandRNG* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstRandSequence* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRange* nodep) { visit(static_cast<AstNodeRange*>(nodep)); }
void VNVisitorConst::visit(AstReadMem* nodep) { visit(static_cast<AstNodeReadWriteMem*>(nodep)); }
void VNVisitorConst::visit(AstRealToBits* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstRedAnd* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstRedOr* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstRedXor* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstRefDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstRelease* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRepeat* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstReplicate* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstReplicateN* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstRequireDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstResizeLValue* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstRestrict* nodep) { visit(static_cast<AstNodeCoverOrAssert*>(nodep)); }
void VNVisitorConst::visit(AstReturn* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstRising* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstRose* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSAnd* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSClocked* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSConsRep* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSEventually* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSExpr* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSFormat* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstSFormatArg* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSFormatF* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSGotoRep* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSIntersect* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSNonConsRep* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSOr* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSScanF* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSThroughout* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSWithin* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSampleQueueDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstSampled* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstScope* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstScopeName* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSel* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSelBit* nodep) { visit(static_cast<AstNodePreSel*>(nodep)); }
void VNVisitorConst::visit(AstSelExtract* nodep) { visit(static_cast<AstNodePreSel*>(nodep)); }
void VNVisitorConst::visit(AstSelMinus* nodep) { visit(static_cast<AstNodePreSel*>(nodep)); }
void VNVisitorConst::visit(AstSelPlus* nodep) { visit(static_cast<AstNodePreSel*>(nodep)); }
void VNVisitorConst::visit(AstSenItem* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstSenTree* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstSequence* nodep) { visit(static_cast<AstNodeFTask*>(nodep)); }
void VNVisitorConst::visit(AstSetAssoc* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSetWildcard* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSetuphold* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstShiftL* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstShiftLOvr* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstShiftR* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstShiftROvr* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstShiftRS* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstShiftRSOvr* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSigned* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstSinD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstSinhD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstSliceSel* nodep) { visit(static_cast<AstNodeTriop*>(nodep)); }
void VNVisitorConst::visit(AstSplitPlaceholder* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstSqrtD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstStable* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstStackTraceF* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstStackTraceT* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstSteady* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstStmtExpr* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstStmtPragma* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstStop* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstStreamDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstStreamL* nodep) { visit(static_cast<AstNodeStream*>(nodep)); }
void VNVisitorConst::visit(AstStreamR* nodep) { visit(static_cast<AstNodeStream*>(nodep)); }
void VNVisitorConst::visit(AstStrengthSpec* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstStructDType* nodep) { visit(static_cast<AstNodeUOrStructDType*>(nodep)); }
void VNVisitorConst::visit(AstStructSel* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSub* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSubD* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstSubstrN* nodep) { visit(static_cast<AstNodeTriop*>(nodep)); }
void VNVisitorConst::visit(AstSysIgnore* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSystemCSection* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstSystemF* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstSystemT* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstTaggedExpr* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstTaggedPattern* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstTanD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstTanhD* nodep) { visit(static_cast<AstNodeSystemUniopD*>(nodep)); }
void VNVisitorConst::visit(AstTask* nodep) { visit(static_cast<AstNodeFTask*>(nodep)); }
void VNVisitorConst::visit(AstTaskRef* nodep) { visit(static_cast<AstNodeFTaskRef*>(nodep)); }
void VNVisitorConst::visit(AstTestPlusArgs* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstText* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstTextBlock* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstThisRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstTime* nodep) { visit(static_cast<AstNodeTermop*>(nodep)); }
void VNVisitorConst::visit(AstTimeD* nodep) { visit(static_cast<AstNodeTermop*>(nodep)); }
void VNVisitorConst::visit(AstTimeFormat* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstTimeImport* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstTimePrecision* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstTimeUnit* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstToLowerN* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstToStringN* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstToUpperN* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstTopScope* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstTraceDecl* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstTraceInc* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstTracePopPrefix* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstTracePushPrefix* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstTypeTable* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstTypedef* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstTypedefFwd* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstURandomRange* nodep) { visit(static_cast<AstNodeBiop*>(nodep)); }
void VNVisitorConst::visit(AstUdpTable* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstUdpTableLine* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstUdpTableLineVal* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstUnbounded* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstUnionDType* nodep) { visit(static_cast<AstNodeUOrStructDType*>(nodep)); }
void VNVisitorConst::visit(AstUnlinkedRef* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstUnpackArrayDType* nodep) { visit(static_cast<AstNodeArrayDType*>(nodep)); }
void VNVisitorConst::visit(AstUnsigned* nodep) { visit(static_cast<AstNodeUniop*>(nodep)); }
void VNVisitorConst::visit(AstUnsizedArrayDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstUnsizedRange* nodep) { visit(static_cast<AstNodeRange*>(nodep)); }
void VNVisitorConst::visit(AstUntil* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstVFile* nodep) { visit(static_cast<AstNodeFile*>(nodep)); }
void VNVisitorConst::visit(AstValuePlusArgs* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstVar* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstVarRef* nodep) { visit(static_cast<AstNodeVarRef*>(nodep)); }
void VNVisitorConst::visit(AstVarScope* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstVarXRef* nodep) { visit(static_cast<AstNodeVarRef*>(nodep)); }
void VNVisitorConst::visit(AstVoidDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstWait* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstWaitFork* nodep) { visit(static_cast<AstNodeStmt*>(nodep)); }
void VNVisitorConst::visit(AstWildcardArrayDType* nodep) { visit(static_cast<AstNodeDType*>(nodep)); }
void VNVisitorConst::visit(AstWildcardRange* nodep) { visit(static_cast<AstNodeRange*>(nodep)); }
void VNVisitorConst::visit(AstWildcardSel* nodep) { visit(static_cast<AstNodeSel*>(nodep)); }
void VNVisitorConst::visit(AstWith* nodep) { visit(static_cast<AstNode*>(nodep)); }
void VNVisitorConst::visit(AstWithParse* nodep) { visit(static_cast<AstNodeExpr*>(nodep)); }
void VNVisitorConst::visit(AstWordSel* nodep) { visit(static_cast<AstNodeSel*>(nodep)); }
void VNVisitorConst::visit(AstWriteMem* nodep) { visit(static_cast<AstNodeReadWriteMem*>(nodep)); }
void VNVisitorConst::visit(AstXor* nodep) { visit(static_cast<AstNodeBiComAsv*>(nodep)); }
