// Generated by astgen // -*- mode: C++; c-file-style: "cc-mode" -*-
    enum en : uint16_t {
        Active = 0,
        Alias = 1,
        AliasScope = 2,
        Arg = 3,
        Bind = 4,
        CFunc = 5,
        CLocalScope = 6,
        CUse = 7,
        CaseItem = 8,
        Cell = 9,
        CellInline = 10,
        CellInlineScope = 11,
        CgOptionAssign = 12,
        ClassExtends = 13,
        Clocking = 14,
        ClockingItem = 15,
        Config = 16,
        ConfigCell = 17,
        ConfigRule = 18,
        ConfigUse = 19,
        ConstPool = 20,
        Constraint = 21,
        ConstraintBefore = 22,
        CoverBin = 23,
        CoverOption = 24,
        CoverTransItem = 25,
        CoverTransSet = 26,
        Covergroup = 27,
        CoverpointRef = 28,
        DefParam = 29,
        DefaultDisable = 30,
        DpiExport = 31,
        ElabDisplay = 32,
        Empty = 33,
        EnumItem = 34,
        ForeachHeader = 35,
        GenCaseItem = 36,
        Implicit = 37,
        InitItem = 38,
        IntfRef = 39,
        Library = 40,
        Modport = 41,
        ModportClockingRef = 42,
        ModportFTaskRef = 43,
        ModportVarRef = 44,
        Netlist = 45,
        PackageExport = 46,
        PackageExportStarStar = 47,
        PackageImport = 48,
        Pin = 49,
        Port = 50,
        Pragma = 51,
        PropSpec = 52,
        Pull = 53,
        Scope = 54,
        SenItem = 55,
        SenTree = 56,
        StrengthSpec = 57,
        SystemCSection = 58,
        Text = 59,
        TextBlock = 60,
        TopScope = 61,
        TypeTable = 62,
        Typedef = 63,
        TypedefFwd = 64,
        UdpTable = 65,
        UdpTableLine = 66,
        UdpTableLineVal = 67,
        Var = 68,
        VarScope = 69,
        With = 70,
        CoverOtherDecl = 71,
        CoverToggleDecl = 72,
        AssocArrayDType = 73,
        BasicDType = 74,
        BracketArrayDType = 75,
        CDType = 76,
        ClassRefDType = 77,
        ConstDType = 78,
        ConstraintRefDType = 79,
        DefImplicitDType = 80,
        DynArrayDType = 81,
        EmptyQueueDType = 82,
        EnumDType = 83,
        IfaceGenericDType = 84,
        IfaceRefDType = 85,
        MemberDType = 86,
        NBACommitQueueDType = 87,
        ParamTypeDType = 88,
        ParseTypeDType = 89,
        QueueDType = 90,
        RefDType = 91,
        RequireDType = 92,
        SampleQueueDType = 93,
        StreamDType = 94,
        UnsizedArrayDType = 95,
        VoidDType = 96,
        WildcardArrayDType = 97,
        PackArrayDType = 98,
        UnpackArrayDType = 99,
        StructDType = 100,
        UnionDType = 101,
        AbortOn = 102,
        AddrOfCFunc = 103,
        AttrOf = 104,
        CExpr = 105,
        CExprUser = 106,
        CMethodHard = 107,
        CReset = 108,
        Cast = 109,
        CastParse = 110,
        CastSize = 111,
        CellArrayRef = 112,
        CellRef = 113,
        ClassOrPackageRef = 114,
        ConsAssoc = 115,
        ConsDynArray = 116,
        ConsPackMember = 117,
        ConsPackUOrStruct = 118,
        ConsQueue = 119,
        ConsWildcard = 120,
        Const = 121,
        ConstraintRef = 122,
        CvtArrayToArray = 123,
        CvtArrayToPacked = 124,
        CvtPackedToArray = 125,
        CvtUnpackedToQueue = 126,
        Dist = 127,
        DistItem = 128,
        Dot = 129,
        EmptyQueue = 130,
        EnumItemRef = 131,
        ExprStmt = 132,
        FError = 133,
        FOpen = 134,
        FOpenMcd = 135,
        FRead = 136,
        FRewind = 137,
        FScanF = 138,
        FSeek = 139,
        FTell = 140,
        Falling = 141,
        Fell = 142,
        Future = 143,
        GatePin = 144,
        GetInitialRandomSeed = 145,
        Implication = 146,
        InitArray = 147,
        Inside = 148,
        InsideRange = 149,
        LambdaArgRef = 150,
        MatchMasked = 151,
        Matches = 152,
        MemberSel = 153,
        NewCopy = 154,
        NewDynamic = 155,
        PExpr = 156,
        ParseHolder = 157,
        ParseRef = 158,
        Past = 159,
        PatMember = 160,
        Pattern = 161,
        PatternStar = 162,
        PatternVar = 163,
        PropAlways = 164,
        Rand = 165,
        RandRNG = 166,
        Rising = 167,
        Rose = 168,
        SClocked = 169,
        SConsRep = 170,
        SEventually = 171,
        SExpr = 172,
        SFormatArg = 173,
        SFormatF = 174,
        SGotoRep = 175,
        SNonConsRep = 176,
        SScanF = 177,
        Sampled = 178,
        ScopeName = 179,
        SetAssoc = 180,
        SetWildcard = 181,
        Stable = 182,
        StackTraceF = 183,
        Steady = 184,
        StructSel = 185,
        SysIgnore = 186,
        SystemF = 187,
        TaggedExpr = 188,
        TaggedPattern = 189,
        TestPlusArgs = 190,
        ThisRef = 191,
        TimePrecision = 192,
        TimeUnit = 193,
        Unbounded = 194,
        UnlinkedRef = 195,
        Until = 196,
        ValuePlusArgs = 197,
        WithParse = 198,
        BufIf1 = 199,
        CastDynamic = 200,
        CompareNN = 201,
        Concat = 202,
        ConcatN = 203,
        Div = 204,
        DivD = 205,
        DivS = 206,
        EqWild = 207,
        FGetS = 208,
        FUngetC = 209,
        GetcN = 210,
        GetcRefN = 211,
        Gt = 212,
        GtD = 213,
        GtN = 214,
        GtS = 215,
        Gte = 216,
        GteD = 217,
        GteN = 218,
        GteS = 219,
        LogAnd = 220,
        LogIf = 221,
        LogOr = 222,
        Lt = 223,
        LtD = 224,
        LtN = 225,
        LtS = 226,
        Lte = 227,
        LteD = 228,
        LteN = 229,
        LteS = 230,
        ModDiv = 231,
        ModDivS = 232,
        NeqWild = 233,
        Pow = 234,
        PowD = 235,
        PowSS = 236,
        PowSU = 237,
        PowUS = 238,
        Replicate = 239,
        ReplicateN = 240,
        SAnd = 241,
        SIntersect = 242,
        SOr = 243,
        SThroughout = 244,
        SWithin = 245,
        Sel = 246,
        ShiftL = 247,
        ShiftLOvr = 248,
        ShiftR = 249,
        ShiftROvr = 250,
        ShiftRS = 251,
        ShiftRSOvr = 252,
        Sub = 253,
        SubD = 254,
        URandomRange = 255,
        Eq = 256,
        EqCase = 257,
        EqD = 258,
        EqN = 259,
        EqT = 260,
        LogEq = 261,
        Neq = 262,
        NeqCase = 263,
        NeqD = 264,
        NeqN = 265,
        NeqT = 266,
        Add = 267,
        AddD = 268,
        And = 269,
        Mul = 270,
        MulD = 271,
        MulS = 272,
        Or = 273,
        Xor = 274,
        DistChiSquare = 275,
        DistExponential = 276,
        DistPoisson = 277,
        DistT = 278,
        ArraySel = 279,
        AssocSel = 280,
        WildcardSel = 281,
        WordSel = 282,
        StreamL = 283,
        StreamR = 284,
        Atan2D = 285,
        HypotD = 286,
        CCall = 287,
        CMethodCall = 288,
        CNew = 289,
        FuncRef = 290,
        MethodCall = 291,
        New = 292,
        TaskRef = 293,
        SelBit = 294,
        SelExtract = 295,
        SelMinus = 296,
        SelPlus = 297,
        CountBits = 298,
        InferredDisable = 299,
        Time = 300,
        TimeD = 301,
        Cond = 302,
        PutcN = 303,
        SliceSel = 304,
        SubstrN = 305,
        DistErlang = 306,
        DistNormal = 307,
        DistUniform = 308,
        AtoN = 309,
        BitsToRealD = 310,
        CCast = 311,
        CLog2 = 312,
        CastWrap = 313,
        CountOnes = 314,
        CvtPackString = 315,
        Extend = 316,
        ExtendS = 317,
        FEof = 318,
        FGetC = 319,
        ISToRD = 320,
        IToRD = 321,
        IsUnbounded = 322,
        IsUnknown = 323,
        LenN = 324,
        LogNot = 325,
        MostSetBitP1 = 326,
        NToI = 327,
        Negate = 328,
        NegateD = 329,
        Not = 330,
        NullCheck = 331,
        OneHot = 332,
        OneHot0 = 333,
        PostDec = 334,
        PostInc = 335,
        PreDec = 336,
        PreInc = 337,
        RToIRoundS = 338,
        RToIS = 339,
        RealToBits = 340,
        RedAnd = 341,
        RedOr = 342,
        RedXor = 343,
        ResizeLValue = 344,
        Signed = 345,
        TimeImport = 346,
        ToLowerN = 347,
        ToStringN = 348,
        ToUpperN = 349,
        Unsigned = 350,
        AcosD = 351,
        AcoshD = 352,
        AsinD = 353,
        AsinhD = 354,
        AtanD = 355,
        AtanhD = 356,
        CeilD = 357,
        CosD = 358,
        CoshD = 359,
        ExpD = 360,
        FloorD = 361,
        Log10D = 362,
        LogD = 363,
        SinD = 364,
        SinhD = 365,
        SqrtD = 366,
        TanD = 367,
        TanhD = 368,
        VarRef = 369,
        VarXRef = 370,
        Func = 371,
        Let = 372,
        Property = 373,
        Sequence = 374,
        Task = 375,
        CFile = 376,
        VFile = 377,
        CoverCross = 378,
        Coverpoint = 379,
        GenBlock = 380,
        GenCase = 381,
        GenFor = 382,
        GenIf = 383,
        Class = 384,
        ClassPackage = 385,
        Iface = 386,
        Module = 387,
        NotFoundModule = 388,
        Package = 389,
        Primitive = 390,
        Always = 391,
        AlwaysObserved = 392,
        AlwaysPost = 393,
        AlwaysPostponed = 394,
        AlwaysPre = 395,
        AlwaysReactive = 396,
        Final = 397,
        Initial = 398,
        InitialAutomatic = 399,
        InitialStatic = 400,
        BracketRange = 401,
        Range = 402,
        UnsizedRange = 403,
        WildcardRange = 404,
        AssertCtl = 405,
        Break = 406,
        CAwait = 407,
        CReturn = 408,
        CStmt = 409,
        CStmtUser = 410,
        Case = 411,
        Comment = 412,
        ConstraintExpr = 413,
        ConstraintUnique = 414,
        Continue = 415,
        CoverInc = 416,
        CoverToggle = 417,
        Deassign = 418,
        Delay = 419,
        Disable = 420,
        DisableFork = 421,
        Display = 422,
        DumpCtl = 423,
        EventControl = 424,
        ExecGraph = 425,
        FClose = 426,
        FFlush = 427,
        Finish = 428,
        FinishFork = 429,
        FireEvent = 430,
        InitialAutomaticStmt = 431,
        InitialStaticStmt = 432,
        JumpBlock = 433,
        JumpGo = 434,
        Loop = 435,
        LoopTest = 436,
        MonitorOff = 437,
        PExprClause = 438,
        PrintTimeScale = 439,
        RSBreak = 440,
        RSProd = 441,
        RSProdItem = 442,
        RSProdList = 443,
        RSReturn = 444,
        RSRule = 445,
        RandCase = 446,
        RandSequence = 447,
        Release = 448,
        Repeat = 449,
        Return = 450,
        SFormat = 451,
        Setuphold = 452,
        SplitPlaceholder = 453,
        StackTraceT = 454,
        StmtExpr = 455,
        StmtPragma = 456,
        Stop = 457,
        SystemT = 458,
        TimeFormat = 459,
        TraceDecl = 460,
        TraceInc = 461,
        TracePopPrefix = 462,
        TracePushPrefix = 463,
        Wait = 464,
        WaitFork = 465,
        Assign = 466,
        AssignCompound = 467,
        AssignCont = 468,
        AssignDly = 469,
        AssignForce = 470,
        AssignW = 471,
        Begin = 472,
        Fork = 473,
        Assert = 474,
        AssertIntrinsic = 475,
        Cover = 476,
        Restrict = 477,
        ConstraintForeach = 478,
        Foreach = 479,
        ConstraintIf = 480,
        If = 481,
        ReadMem = 482,
        WriteMem = 483,
    };
    static constexpr size_t NUM_TYPES() { return 484; }
    enum bounds : uint16_t {
        firstNode = 0,
        lastNode  = 483,
        firstNodeCoverDecl = 71,
        lastNodeCoverDecl  = 72,
        firstNodeDType = 73,
        lastNodeDType  = 101,
        firstNodeArrayDType = 98,
        lastNodeArrayDType  = 99,
        firstNodeUOrStructDType = 100,
        lastNodeUOrStructDType  = 101,
        firstNodeExpr = 102,
        lastNodeExpr  = 370,
        firstNodeBiop = 199,
        lastNodeBiop  = 286,
        firstNodeBiCom = 256,
        lastNodeBiCom  = 274,
        firstNodeBiComAsv = 267,
        lastNodeBiComAsv  = 274,
        firstNodeDistBiop = 275,
        lastNodeDistBiop  = 278,
        firstNodeSel = 279,
        lastNodeSel  = 282,
        firstNodeStream = 283,
        lastNodeStream  = 284,
        firstNodeSystemBiopD = 285,
        lastNodeSystemBiopD  = 286,
        firstNodeCCall = 287,
        lastNodeCCall  = 289,
        firstNodeFTaskRef = 290,
        lastNodeFTaskRef  = 293,
        firstNodePreSel = 294,
        lastNodePreSel  = 297,
        firstNodeQuadop = 298,
        lastNodeQuadop  = 298,
        firstNodeTermop = 299,
        lastNodeTermop  = 301,
        firstNodeTriop = 302,
        lastNodeTriop  = 308,
        firstNodeDistTriop = 306,
        lastNodeDistTriop  = 308,
        firstNodeUniop = 309,
        lastNodeUniop  = 368,
        firstNodeSystemUniopD = 351,
        lastNodeSystemUniopD  = 368,
        firstNodeVarRef = 369,
        lastNodeVarRef  = 370,
        firstNodeFTask = 371,
        lastNodeFTask  = 375,
        firstNodeFile = 376,
        lastNodeFile  = 377,
        firstNodeFuncCovItem = 378,
        lastNodeFuncCovItem  = 379,
        firstNodeGen = 380,
        lastNodeGen  = 383,
        firstNodeModule = 384,
        lastNodeModule  = 390,
        firstNodeProcedure = 391,
        lastNodeProcedure  = 400,
        firstNodeRange = 401,
        lastNodeRange  = 404,
        firstNodeStmt = 405,
        lastNodeStmt  = 483,
        firstNodeAssign = 466,
        lastNodeAssign  = 471,
        firstNodeBlock = 472,
        lastNodeBlock  = 473,
        firstNodeCoverOrAssert = 474,
        lastNodeCoverOrAssert  = 477,
        firstNodeForeach = 478,
        lastNodeForeach  = 479,
        firstNodeIf = 480,
        lastNodeIf  = 481,
        firstNodeReadWriteMem = 482,
        lastNodeReadWriteMem  = 483,
    };
    const char* ascii() const VL_MT_SAFE {
        static const char* const names[] = {
            "ACTIVE",
            "ALIAS",
            "ALIASSCOPE",
            "ARG",
            "BIND",
            "CFUNC",
            "CLOCALSCOPE",
            "CUSE",
            "CASEITEM",
            "CELL",
            "CELLINLINE",
            "CELLINLINESCOPE",
            "CGOPTIONASSIGN",
            "CLASSEXTENDS",
            "CLOCKING",
            "CLOCKINGITEM",
            "CONFIG",
            "CONFIGCELL",
            "CONFIGRULE",
            "CONFIGUSE",
            "CONSTPOOL",
            "CONSTRAINT",
            "CONSTRAINTBEFORE",
            "COVERBIN",
            "COVEROPTION",
            "COVERTRANSITEM",
            "COVERTRANSSET",
            "COVERGROUP",
            "COVERPOINTREF",
            "DEFPARAM",
            "DEFAULTDISABLE",
            "DPIEXPORT",
            "ELABDISPLAY",
            "EMPTY",
            "ENUMITEM",
            "FOREACHHEADER",
            "GENCASEITEM",
            "IMPLICIT",
            "INITITEM",
            "INTFREF",
            "LIBRARY",
            "MODPORT",
            "MODPORTCLOCKINGREF",
            "MODPORTFTASKREF",
            "MODPORTVARREF",
            "NETLIST",
            "PACKAGEEXPORT",
            "PACKAGEEXPORTSTARSTAR",
            "PACKAGEIMPORT",
            "PIN",
            "PORT",
            "PRAGMA",
            "PROPSPEC",
            "PULL",
            "SCOPE",
            "SENITEM",
            "SENTREE",
            "STRENGTHSPEC",
            "SYSTEMCSECTION",
            "TEXT",
            "TEXTBLOCK",
            "TOPSCOPE",
            "TYPETABLE",
            "TYPEDEF",
            "TYPEDEFFWD",
            "UDPTABLE",
            "UDPTABLELINE",
            "UDPTABLELINEVAL",
            "VAR",
            "VARSCOPE",
            "WITH",
            "COVEROTHERDECL",
            "COVERTOGGLEDECL",
            "ASSOCARRAYDTYPE",
            "BASICDTYPE",
            "BRACKETARRAYDTYPE",
            "CDTYPE",
            "CLASSREFDTYPE",
            "CONSTDTYPE",
            "CONSTRAINTREFDTYPE",
            "DEFIMPLICITDTYPE",
            "DYNARRAYDTYPE",
            "EMPTYQUEUEDTYPE",
            "ENUMDTYPE",
            "IFACEGENERICDTYPE",
            "IFACEREFDTYPE",
            "MEMBERDTYPE",
            "NBACOMMITQUEUEDTYPE",
            "PARAMTYPEDTYPE",
            "PARSETYPEDTYPE",
            "QUEUEDTYPE",
            "REFDTYPE",
            "REQUIREDTYPE",
            "SAMPLEQUEUEDTYPE",
            "STREAMDTYPE",
            "UNSIZEDARRAYDTYPE",
            "VOIDDTYPE",
            "WILDCARDARRAYDTYPE",
            "PACKARRAYDTYPE",
            "UNPACKARRAYDTYPE",
            "STRUCTDTYPE",
            "UNIONDTYPE",
            "ABORTON",
            "ADDROFCFUNC",
            "ATTROF",
            "CEXPR",
            "CEXPRUSER",
            "CMETHODHARD",
            "CRESET",
            "CAST",
            "CASTPARSE",
            "CASTSIZE",
            "CELLARRAYREF",
            "CELLREF",
            "CLASSORPACKAGEREF",
            "CONSASSOC",
            "CONSDYNARRAY",
            "CONSPACKMEMBER",
            "CONSPACKUORSTRUCT",
            "CONSQUEUE",
            "CONSWILDCARD",
            "CONST",
            "CONSTRAINTREF",
            "CVTARRAYTOARRAY",
            "CVTARRAYTOPACKED",
            "CVTPACKEDTOARRAY",
            "CVTUNPACKEDTOQUEUE",
            "DIST",
            "DISTITEM",
            "DOT",
            "EMPTYQUEUE",
            "ENUMITEMREF",
            "EXPRSTMT",
            "FERROR",
            "FOPEN",
            "FOPENMCD",
            "FREAD",
            "FREWIND",
            "FSCANF",
            "FSEEK",
            "FTELL",
            "FALLING",
            "FELL",
            "FUTURE",
            "GATEPIN",
            "GETINITIALRANDOMSEED",
            "IMPLICATION",
            "INITARRAY",
            "INSIDE",
            "INSIDERANGE",
            "LAMBDAARGREF",
            "MATCHMASKED",
            "MATCHES",
            "MEMBERSEL",
            "NEWCOPY",
            "NEWDYNAMIC",
            "PEXPR",
            "PARSEHOLDER",
            "PARSEREF",
            "PAST",
            "PATMEMBER",
            "PATTERN",
            "PATTERNSTAR",
            "PATTERNVAR",
            "PROPALWAYS",
            "RAND",
            "RANDRNG",
            "RISING",
            "ROSE",
            "SCLOCKED",
            "SCONSREP",
            "SEVENTUALLY",
            "SEXPR",
            "SFORMATARG",
            "SFORMATF",
            "SGOTOREP",
            "SNONCONSREP",
            "SSCANF",
            "SAMPLED",
            "SCOPENAME",
            "SETASSOC",
            "SETWILDCARD",
            "STABLE",
            "STACKTRACEF",
            "STEADY",
            "STRUCTSEL",
            "SYSIGNORE",
            "SYSTEMF",
            "TAGGEDEXPR",
            "TAGGEDPATTERN",
            "TESTPLUSARGS",
            "THISREF",
            "TIMEPRECISION",
            "TIMEUNIT",
            "UNBOUNDED",
            "UNLINKEDREF",
            "UNTIL",
            "VALUEPLUSARGS",
            "WITHPARSE",
            "BUFIF1",
            "CASTDYNAMIC",
            "COMPARENN",
            "CONCAT",
            "CONCATN",
            "DIV",
            "DIVD",
            "DIVS",
            "EQWILD",
            "FGETS",
            "FUNGETC",
            "GETCN",
            "GETCREFN",
            "GT",
            "GTD",
            "GTN",
            "GTS",
            "GTE",
            "GTED",
            "GTEN",
            "GTES",
            "LOGAND",
            "LOGIF",
            "LOGOR",
            "LT",
            "LTD",
            "LTN",
            "LTS",
            "LTE",
            "LTED",
            "LTEN",
            "LTES",
            "MODDIV",
            "MODDIVS",
            "NEQWILD",
            "POW",
            "POWD",
            "POWSS",
            "POWSU",
            "POWUS",
            "REPLICATE",
            "REPLICATEN",
            "SAND",
            "SINTERSECT",
            "SOR",
            "STHROUGHOUT",
            "SWITHIN",
            "SEL",
            "SHIFTL",
            "SHIFTLOVR",
            "SHIFTR",
            "SHIFTROVR",
            "SHIFTRS",
            "SHIFTRSOVR",
            "SUB",
            "SUBD",
            "URANDOMRANGE",
            "EQ",
            "EQCASE",
            "EQD",
            "EQN",
            "EQT",
            "LOGEQ",
            "NEQ",
            "NEQCASE",
            "NEQD",
            "NEQN",
            "NEQT",
            "ADD",
            "ADDD",
            "AND",
            "MUL",
            "MULD",
            "MULS",
            "OR",
            "XOR",
            "DISTCHISQUARE",
            "DISTEXPONENTIAL",
            "DISTPOISSON",
            "DISTT",
            "ARRAYSEL",
            "ASSOCSEL",
            "WILDCARDSEL",
            "WORDSEL",
            "STREAML",
            "STREAMR",
            "ATAN2D",
            "HYPOTD",
            "CCALL",
            "CMETHODCALL",
            "CNEW",
            "FUNCREF",
            "METHODCALL",
            "NEW",
            "TASKREF",
            "SELBIT",
            "SELEXTRACT",
            "SELMINUS",
            "SELPLUS",
            "COUNTBITS",
            "INFERREDDISABLE",
            "TIME",
            "TIMED",
            "COND",
            "PUTCN",
            "SLICESEL",
            "SUBSTRN",
            "DISTERLANG",
            "DISTNORMAL",
            "DISTUNIFORM",
            "ATON",
            "BITSTOREALD",
            "CCAST",
            "CLOG2",
            "CASTWRAP",
            "COUNTONES",
            "CVTPACKSTRING",
            "EXTEND",
            "EXTENDS",
            "FEOF",
            "FGETC",
            "ISTORD",
            "ITORD",
            "ISUNBOUNDED",
            "ISUNKNOWN",
            "LENN",
            "LOGNOT",
            "MOSTSETBITP1",
            "NTOI",
            "NEGATE",
            "NEGATED",
            "NOT",
            "NULLCHECK",
            "ONEHOT",
            "ONEHOT0",
            "POSTDEC",
            "POSTINC",
            "PREDEC",
            "PREINC",
            "RTOIROUNDS",
            "RTOIS",
            "REALTOBITS",
            "REDAND",
            "REDOR",
            "REDXOR",
            "RESIZELVALUE",
            "SIGNED",
            "TIMEIMPORT",
            "TOLOWERN",
            "TOSTRINGN",
            "TOUPPERN",
            "UNSIGNED",
            "ACOSD",
            "ACOSHD",
            "ASIND",
            "ASINHD",
            "ATAND",
            "ATANHD",
            "CEILD",
            "COSD",
            "COSHD",
            "EXPD",
            "FLOORD",
            "LOG10D",
            "LOGD",
            "SIND",
            "SINHD",
            "SQRTD",
            "TAND",
            "TANHD",
            "VARREF",
            "VARXREF",
            "FUNC",
            "LET",
            "PROPERTY",
            "SEQUENCE",
            "TASK",
            "CFILE",
            "VFILE",
            "COVERCROSS",
            "COVERPOINT",
            "GENBLOCK",
            "GENCASE",
            "GENFOR",
            "GENIF",
            "CLASS",
            "CLASSPACKAGE",
            "IFACE",
            "MODULE",
            "NOTFOUNDMODULE",
            "PACKAGE",
            "PRIMITIVE",
            "ALWAYS",
            "ALWAYSOBSERVED",
            "ALWAYSPOST",
            "ALWAYSPOSTPONED",
            "ALWAYSPRE",
            "ALWAYSREACTIVE",
            "FINAL",
            "INITIAL",
            "INITIALAUTOMATIC",
            "INITIALSTATIC",
            "BRACKETRANGE",
            "RANGE",
            "UNSIZEDRANGE",
            "WILDCARDRANGE",
            "ASSERTCTL",
            "BREAK",
            "CAWAIT",
            "CRETURN",
            "CSTMT",
            "CSTMTUSER",
            "CASE",
            "COMMENT",
            "CONSTRAINTEXPR",
            "CONSTRAINTUNIQUE",
            "CONTINUE",
            "COVERINC",
            "COVERTOGGLE",
            "DEASSIGN",
            "DELAY",
            "DISABLE",
            "DISABLEFORK",
            "DISPLAY",
            "DUMPCTL",
            "EVENTCONTROL",
            "EXECGRAPH",
            "FCLOSE",
            "FFLUSH",
            "FINISH",
            "FINISHFORK",
            "FIREEVENT",
            "INITIALAUTOMATICSTMT",
            "INITIALSTATICSTMT",
            "JUMPBLOCK",
            "JUMPGO",
            "LOOP",
            "LOOPTEST",
            "MONITOROFF",
            "PEXPRCLAUSE",
            "PRINTTIMESCALE",
            "RSBREAK",
            "RSPROD",
            "RSPRODITEM",
            "RSPRODLIST",
            "RSRETURN",
            "RSRULE",
            "RANDCASE",
            "RANDSEQUENCE",
            "RELEASE",
            "REPEAT",
            "RETURN",
            "SFORMAT",
            "SETUPHOLD",
            "SPLITPLACEHOLDER",
            "STACKTRACET",
            "STMTEXPR",
            "STMTPRAGMA",
            "STOP",
            "SYSTEMT",
            "TIMEFORMAT",
            "TRACEDECL",
            "TRACEINC",
            "TRACEPOPPREFIX",
            "TRACEPUSHPREFIX",
            "WAIT",
            "WAITFORK",
            "ASSIGN",
            "ASSIGNCOMPOUND",
            "ASSIGNCONT",
            "ASSIGNDLY",
            "ASSIGNFORCE",
            "ASSIGNW",
            "BEGIN",
            "FORK",
            "ASSERT",
            "ASSERTINTRINSIC",
            "COVER",
            "RESTRICT",
            "CONSTRAINTFOREACH",
            "FOREACH",
            "CONSTRAINTIF",
            "IF",
            "READMEM",
            "WRITEMEM",
        };
        return names[m_e];
    }
